"""A 1D anisotropic grating example.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

from typing import Tuple

import jax.numpy as jnp

from fmmax import basis, fields, fmm, layer, scattering, utils

WAVELENGTH_NM: jnp.ndarray = jnp.array([450.0, 550.0, 620.0])
POLAR_ANGLE: jnp.ndarray = jnp.array([0.0])
AZIMUTHAL_ANGLE: jnp.ndarray = jnp.array([0.0])


def simulate_grating(
    permittivity_ambient: complex = 1.0 + 0.0j,
    wavelength_nm: jnp.ndarray = WAVELENGTH_NM,
    polar_angle: jnp.ndarray = POLAR_ANGLE,
    azimuthal_angle: jnp.ndarray = AZIMUTHAL_ANGLE,
    pitch_nm: float = 180.0,
    grating_width_nm: float = 60.0,
    grating_thickness_nm: float = 80.0,
    resolution_nm: float = 1.0,
    approximate_num_terms: int = 10,
    truncation: basis.Truncation = basis.Truncation.CIRCULAR,
) -> Tuple[int, jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Computes the TE- and TM-polarized reflection from a 1D stripe grating.

    Args:
        permittivity_ambient: The permittivity of the ambient.
        wavelength_nm: The excitation wavelength, in nanometers.
        polar_angle: The polar angle of the incident plane wave.
        azimuthal_angle: The azimuthal angle of the incident plane wave.
        pitch_nm: The grating pitch, in nanometers.
        grating_width_nm: The width of the lines comprising the grating.
        grating_thickness_nm: The height of the grating.
        resolution_nm: The rasterization resolution for patterned layers.
        approximate_num_terms: The approximate number of terms used in the plane
            wave expansion of the fields.
        truncation: Determines the truncation of the expansion.

    Returns:
        The number of terms in the expansion, and the incident, reflected, and
        transmitted power for each order.
    """
    # Add spatial dimensions to the scalar ambient permittivity.
    permittivity_ambient_ = jnp.asarray(permittivity_ambient)[jnp.newaxis, jnp.newaxis]
    del permittivity_ambient

    # Define the permittivity tensor elements for the grating gap.
    permittivities_grating_gap = (
        permittivity_ambient_,  # xx
        jnp.zeros_like(permittivity_ambient_),  # xy
        jnp.zeros_like(permittivity_ambient_),  # yx
        permittivity_ambient_,  # yy
        permittivity_ambient_,  # zz
    )

    # Get the permittivity tensor elements for the anisotropic, and add the spatial dimensions.
    permittivities_anisotropic = permittivity_tensor_elements_anisotropic(wavelength_nm)
    permittivities_grating_tooth = [
        p[..., jnp.newaxis, jnp.newaxis] for p in permittivities_anisotropic
    ]

    # Get the permittivity tensor elements for the grating using interpolation.
    density = grating_density(pitch_nm, grating_width_nm, resolution_nm)
    permittivities_grating = [
        utils.interpolate_permittivity(
            permittivity_solid=pt, permittivity_void=pg, density=density  # type: ignore[arg-type]
        )
        for pt, pg in zip(permittivities_grating_tooth, permittivities_grating_gap)
    ]

    # Add spatial dimensions to the substrate permittivity tensor elements.
    permittivities_substrate = [
        p[..., jnp.newaxis, jnp.newaxis] for p in permittivities_anisotropic
    ]

    # Compute quantities needed for the eigensolve.
    thicknesses = [jnp.zeros(()), jnp.asarray(grating_thickness_nm), jnp.zeros(())]

    # We are modeling a 1D grating with a 2D solver. To do this efficiently, we make the
    # the 2D unit cell very small along the invariant direction, so that grating vectors
    # are well allocated.
    primitive_lattice_vectors = basis.LatticeVectors(
        u=basis.X * pitch_nm, v=basis.Y * pitch_nm / (2 * approximate_num_terms)
    )
    expansion = basis.generate_expansion(
        primitive_lattice_vectors=primitive_lattice_vectors,
        approximate_num_terms=approximate_num_terms,
        truncation=truncation,
    )
    # Assert that grating vectors all have ky == 0.
    assert max(abs(expansion.basis_coefficients[:, 1])) == 0

    in_plane_wavevector = basis.plane_wave_in_plane_wavevector(
        wavelength=wavelength_nm,
        polar_angle=polar_angle,
        azimuthal_angle=azimuthal_angle,
        permittivity=permittivity_ambient_,
    )

    # Perform the appropriate layer eigensolve for each layer in the stack. Note that the
    # ordering of tensor elements in the `permittivities_grating` and `permittivities_substrate`
    # must must match those in the arguments of `eigensolve_patterned_anisotropic_media`.
    layer_solve_results = [
        layer.eigensolve_isotropic_media(
            wavelength=wavelength_nm,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            permittivity=permittivity_ambient_,
            expansion=expansion,
            formulation=fmm.Formulation.FFT,
        ),
        layer.eigensolve_anisotropic_media(
            wavelength_nm,
            in_plane_wavevector,
            primitive_lattice_vectors,
            *permittivities_grating,  # type: ignore[arg-type]
            expansion=expansion,  # type: ignore[misc]
            formulation=fmm.Formulation.FFT,  # type: ignore[misc]
        ),
        layer.eigensolve_anisotropic_media(
            wavelength_nm,
            in_plane_wavevector,
            primitive_lattice_vectors,
            *permittivities_substrate,  # type: ignore[arg-type]
            expansion=expansion,  # type: ignore[misc]
            formulation=fmm.Formulation.FFT,  # type: ignore[misc]
        ),
    ]

    # Compute the scattering matrix for our stack.
    s_matrix = scattering.stack_s_matrix(
        layer_solve_results=layer_solve_results,
        layer_thicknesses=[jnp.asarray(t) for t in thicknesses],
    )

    # Generate the wave amplitudes for the plane excitation. We excite with
    # both TE-polarized and TM-polarized light.
    n = layer_solve_results[0].expansion.num_terms
    forward_amplitude_0_start = jnp.zeros((2 * n, 2), dtype=complex)
    forward_amplitude_0_start = forward_amplitude_0_start.at[0, 0].set(1)  # te
    forward_amplitude_0_start = forward_amplitude_0_start.at[n, 1].set(1)  # tm

    backward_amplitude_0_end = s_matrix.s21 @ forward_amplitude_0_start
    _, backward_amplitude_0_start = fields.colocate_amplitudes(
        forward_amplitude_0_start,
        backward_amplitude_0_end,
        z_offset=jnp.zeros(()),
        layer_solve_result=layer_solve_results[0],
        layer_thickness=jnp.asarray(thicknesses[0]),
    )
    incident_power, reflected_power = fields.amplitude_poynting_flux(
        forward_amplitude=forward_amplitude_0_start,
        backward_amplitude=backward_amplitude_0_start,
        layer_solve_result=layer_solve_results[0],
    )

    forward_amplitude_N_start = s_matrix.s11 @ forward_amplitude_0_start
    transmitted_power, _ = fields.amplitude_poynting_flux(
        forward_amplitude=forward_amplitude_N_start,
        backward_amplitude=jnp.zeros_like(forward_amplitude_N_start),
        layer_solve_result=layer_solve_results[-1],
    )

    return expansion.num_terms, incident_power, reflected_power, transmitted_power


def permittivity_tensor_elements_anisotropic(
    wavelength_nm: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Returns the permittivity tensor elements of an anisotropic material.

    Args:
        wavelength_nm: The wavelength for which the permittivity is sought, with
            units of nanometers.

    Returns:
        The permittivity tensor elements `(exx, exy, eyx, eyy, ezz)`.
    """
    return (
        jnp.ones_like(wavelength_nm) * 2.0**2,
        jnp.zeros_like(wavelength_nm),
        jnp.zeros_like(wavelength_nm),
        jnp.ones_like(wavelength_nm) * 2.4**2,
        jnp.ones_like(wavelength_nm) * 2.8**2,
    )


def grating_density(
    pitch: float,
    grating_width: float,
    resolution: float,
) -> jnp.ndarray:
    """Returns the density for a grating.

    The density is `1` inside the grating tooth, and `0` inside the gap.

    Args:
        pitch: The grating pitch.
        grating_width: The grating width.
        resolution: The resolution of the density array.

    Returns:
        The density array, with shape `(pitch / resolution, 1)`.
    """
    x = jnp.arange(-pitch / 2, pitch / 2, resolution)
    x = x[:, jnp.newaxis]
    return (jnp.abs(x) <= grating_width / 2).astype(float)


if __name__ == "__main__":
    print(simulate_grating())



"""Tests for `fmmax.fmm`.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import unittest

import jax
import jax.numpy as jnp
import numpy as onp
import parameterized

from fmmax import basis, fmm

# Enable 64-bit precision for higher-accuracy.
jax.config.update("jax_enable_x64", True)


PRIMITIVE_LATTICE_VECTORS = basis.LatticeVectors(
    u=jnp.array([1, 0]), v=jnp.array([0, 1])
)
EXPANSION = basis.generate_expansion(
    primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
    approximate_num_terms=30,
    truncation=basis.Truncation.CIRCULAR,
)


class FftTest(unittest.TestCase):
    def test_fft_ifft_single(self):
        shape = (100, 80)
        x = jax.random.uniform(jax.random.PRNGKey(0), shape=shape)

        y = fmm.fft(x, EXPANSION)
        self.assertSequenceEqual(y.shape, (EXPANSION.num_terms,))

        ifft_y = fmm.ifft(y, EXPANSION, shape)
        self.assertSequenceEqual(ifft_y.shape, shape)

        fft_ifft_y = fmm.fft(ifft_y, EXPANSION)
        onp.testing.assert_allclose(y, fft_ifft_y)

    def test_fft_specify_axis(self):
        shape = (20, 100, 80, 4)
        x = jax.random.uniform(jax.random.PRNGKey(0), shape=shape)
        y = fmm.fft(x, EXPANSION, axes=(-3, -2))
        expected_y_transpose = fmm.fft(
            jnp.transpose(x, (0, 3, 1, 2)),
            EXPANSION,
            axes=(-2, -1),
        )
        expected_y = jnp.transpose(expected_y_transpose, (0, 2, 1))
        onp.testing.assert_array_equal(y, expected_y)

    def test_ifft_specify_axis(self):
        shape = (20, 100, 80)
        x = jax.random.uniform(jax.random.PRNGKey(0), shape=shape)
        y = fmm.fft(x, EXPANSION)
        expected_ifft_y = fmm.ifft(y, EXPANSION, shape=(100, 80))
        self.assertSequenceEqual(y.shape, (20, EXPANSION.num_terms))

        y_transposed = jnp.transpose(y, axes=(1, 0))
        ifft_y_transposed = fmm.ifft(y_transposed, EXPANSION, shape=(100, 80), axis=-2)
        ifft_y = jnp.transpose(ifft_y_transposed, axes=(2, 0, 1))
        onp.testing.assert_allclose(ifft_y, expected_ifft_y)

    def test_fft_ifft_batch(self):
        batch_size = 24
        shape = (100, 80)
        x = jax.random.uniform(jax.random.PRNGKey(0), shape=(batch_size,) + shape)

        y = fmm.fft(x, EXPANSION)
        self.assertSequenceEqual(y.shape, (batch_size, EXPANSION.num_terms))

        ifft_y = fmm.ifft(y, EXPANSION, shape)
        self.assertSequenceEqual(ifft_y.shape, (batch_size,) + shape)

        fft_ifft_y = fmm.fft(ifft_y, EXPANSION)
        onp.testing.assert_allclose(y, fft_ifft_y)


class TransversePermittivityTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            (fmm.Formulation.JONES_DIRECT,),
            (fmm.Formulation.JONES,),
            (fmm.Formulation.NORMAL,),
            (fmm.Formulation.POL,),
        ]
    )
    def test_single_matches_batch_vector(self, fmm_formulation):
        x, y = jnp.meshgrid(
            jnp.linspace(-0.5, 0.5),
            jnp.linspace(-0.5, 0.5),
            indexing="ij",
        )
        circle = (jnp.sqrt(x**2 + y**2) <= 0.2).astype(float)
        scale = jnp.arange(1, 5)[:, jnp.newaxis, jnp.newaxis]
        permittivity = 1 + circle * scale
        result_batch = fmm._transverse_permittivity_vector(
            PRIMITIVE_LATTICE_VECTORS, permittivity, EXPANSION, fmm_formulation
        )
        result_single = [
            fmm._transverse_permittivity_vector(
                primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
                permittivity=p,
                expansion=EXPANSION,
                formulation=fmm_formulation,
            )
            for p in permittivity
        ]
        onp.testing.assert_allclose(result_batch, result_single, atol=1e-15)


class ToeplitzIndicesTest(unittest.TestCase):
    def test_standard(self):
        expansion = basis.generate_expansion(
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
            approximate_num_terms=30,
            truncation=basis.Truncation.CIRCULAR,
        )
        indices = fmm._standard_toeplitz_indices(expansion)
        for row_coeff, row in zip(expansion.basis_coefficients, indices):
            for col_coeff, idx in zip(expansion.basis_coefficients, row):
                self.assertSequenceEqual(row_coeff.shape, (2,))
                self.assertSequenceEqual(col_coeff.shape, (2,))
                self.assertSequenceEqual(idx.shape, (2,))
                onp.testing.assert_array_equal(-idx + row_coeff, col_coeff)


class ShapeValidationTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            (onp.asarray([[-2, -2], [2, 2]]), (5, 5)),
            (onp.asarray([[-5, -2], [5, 2]]), (11, 5)),
            (onp.asarray([[0, -2], [8, 3]]), (17, 7)),
        ]
    )
    def test_min_shape(self, basis_coefficients, expected_min_shape):
        expansion = basis.Expansion(basis_coefficients=basis_coefficients)
        min_shape = fmm._min_array_shape_for_expansion(expansion)
        self.assertSequenceEqual(min_shape, expected_min_shape)

    def test_fourier_convolution_matrix(self):
        expansion = basis.Expansion(basis_coefficients=jnp.asarray([[-2, -2], [2, 2]]))
        with self.assertRaisesRegex(ValueError, "`shape` is insufficient for"):
            fmm.fourier_convolution_matrix(jnp.ones((8, 4)), expansion)

    def test_fft(self):
        expansion = basis.Expansion(basis_coefficients=jnp.asarray([[-2, -2], [2, 2]]))
        with self.assertRaisesRegex(ValueError, "`shape` is insufficient for"):
            fmm.fft(jnp.ones((8, 4)), expansion)

    def test_ifft(self):
        expansion = basis.Expansion(basis_coefficients=jnp.asarray([[-2, -2], [2, 2]]))
        with self.assertRaisesRegex(ValueError, "`shape` is insufficient for"):
            fmm.ifft(jnp.ones((2,)), expansion, shape=(8, 4))


class AnistropicLayerTest(unittest.TestCase):
    def test_compare_when_layer_is_isotropic(self):
        permittivity = 1 + jax.random.uniform(jax.random.PRNGKey(0), (50, 50))
        (
            eta_expected,
            z_expected,
            transverse_expected,
        ) = fmm.fourier_matrices_patterned_isotropic_media(
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
            permittivity=permittivity,
            expansion=EXPANSION,
            formulation=fmm.Formulation.FFT,
        )
        (
            eta_result,
            z_result,
            transverse_result,
        ) = fmm.fourier_matrices_patterned_anisotropic_media(
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
            permittivity_xx=permittivity,
            permittivity_xy=jnp.zeros_like(permittivity),
            permittivity_yx=jnp.zeros_like(permittivity),
            permittivity_yy=permittivity,
            permittivity_zz=permittivity,
            expansion=EXPANSION,
            formulation=fmm.Formulation.FFT,
        )
        onp.testing.assert_array_equal(eta_result, eta_expected)
        onp.testing.assert_array_equal(z_result, z_expected)
        onp.testing.assert_array_equal(transverse_result, transverse_expected)



"""Validation tests that involve simulations of various structures.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import unittest

import jax
import jax.numpy as jnp
import numpy as onp
import parameterized

from fmmax import basis, fields, fmm, layer, scattering

# Enable 64-bit precision for higher accuracy.
jax.config.update("jax_enable_x64", True)


def _solve_s_matrix(
    wavelength,
    in_plane_wavevector,
    primitive_lattice_vectors,
    expansion,
    permittivities,
    thicknesses,
    formulation=fmm.Formulation.FFT,
):
    layer_solve_results = [
        layer.eigensolve_isotropic_media(
            wavelength=wavelength,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            permittivity=p,
            expansion=expansion,
            formulation=formulation,
        )
        for p in permittivities
    ]
    s_matrix = scattering.stack_s_matrix(layer_solve_results, thicknesses)
    return s_matrix, layer_solve_results


def _solve_fresnel_reflection(n1, n2, theta_i):
    theta_t = jnp.arcsin(n1 / n2 * jnp.sin(theta_i))
    reflection_s = (
        jnp.abs(
            (n1 * jnp.cos(theta_i) - n2 * jnp.cos(theta_t))
            / (n1 * jnp.cos(theta_i) + n2 * jnp.cos(theta_t))
        )
        ** 2
    )
    reflection_p = (
        jnp.abs(
            (n1 * jnp.cos(theta_t) - n2 * jnp.cos(theta_i))
            / (n1 * jnp.cos(theta_t) + n2 * jnp.cos(theta_i))
        )
        ** 2
    )
    return reflection_s, reflection_p


class ThinFilmComparisonTest(unittest.TestCase):
    @parameterized.parameterized.expand([((1, 1),), ((64, 64),)])
    def test_normal_incidence_matches_fresnel(self, permittivity_shape):
        n1 = 1.2
        n2 = 3.8
        wavelength = jnp.array(0.238)
        in_plane_wavevector = jnp.zeros((2,))
        primitive_lattice_vectors = basis.LatticeVectors(
            u=jnp.asarray([1.0, 0.0]), v=jnp.asarray([0.0, 0.8])
        )
        expansion = basis.generate_expansion(
            primitive_lattice_vectors=primitive_lattice_vectors,
            approximate_num_terms=10,
            truncation=basis.Truncation.CIRCULAR,
        )
        permittivities = [
            jnp.broadcast_to(jnp.asarray([[n1**2]]), permittivity_shape),
            jnp.broadcast_to(jnp.asarray([[n2**2]]), permittivity_shape),
        ]
        thicknesses = [1.0, 2.0]
        s_matrix, _ = _solve_s_matrix(
            wavelength=wavelength,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            expansion=expansion,
            permittivities=permittivities,
            thicknesses=thicknesses,
        )
        # Ensure that the `0`-th basis coefficient is for the zeroth order.
        onp.testing.assert_array_equal(expansion.basis_coefficients[0, :], (0, 0))

        # Extract the zeroth-order reflection.
        reflection = jnp.abs(s_matrix.s12[0, 0]) ** 2
        expected_reflection = jnp.abs((n1 - n2) / (n1 + n2)) ** 2
        onp.testing.assert_allclose(reflection, expected_reflection)
        # Extract the zeroth-order transmission.
        transmission = jnp.abs(s_matrix.s11[0, 0]) ** 2 * n1 / n2
        onp.testing.assert_allclose(transmission, 1 - expected_reflection)

    def test_oblique_incidence_matches_fresenl(self):
        n1 = 1.2
        n2 = 3.8
        wavelength = jnp.array(0.238)
        polar_angle = 0.0
        in_plane_wavevector = basis.plane_wave_in_plane_wavevector(
            wavelength=wavelength,
            polar_angle=polar_angle,
            azimuthal_angle=0.0,
            permittivity=n1**2,
        )
        primitive_lattice_vectors = basis.LatticeVectors(
            u=jnp.asarray([1.0, 0.0]), v=jnp.asarray([0.0, 1.0])
        )
        expansion = basis.generate_expansion(
            primitive_lattice_vectors=primitive_lattice_vectors,
            approximate_num_terms=5,
            truncation=basis.Truncation.CIRCULAR,
        )
        num_terms = expansion.num_terms
        permittivities = [jnp.asarray([[n1**2]]), jnp.asarray([[n2**2]])]
        thicknesses = [1.0, 1.0]

        s_matrix, layer_solve_results = _solve_s_matrix(
            wavelength=wavelength,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            expansion=expansion,
            permittivities=permittivities,
            thicknesses=thicknesses,
        )

        expected_reflection_s, expected_reflection_p = _solve_fresnel_reflection(
            n1, n2, polar_angle
        )

        # Check that reflection matches expected, using the s-matrix directly.
        with self.subTest("reflection from s-matrix"):
            reflection_s = jnp.abs(s_matrix.s21[0, 0]) ** 2
            reflection_p = jnp.abs(s_matrix.s21[num_terms, num_terms]) ** 2
            onp.testing.assert_allclose(reflection_s, expected_reflection_s)
            onp.testing.assert_allclose(reflection_p, expected_reflection_p)

        # Check that reflection matches expected, using the poynting vector.
        with self.subTest("reflection from poynting flux"):
            forward_amplitude_0_start = jnp.ones((2 * num_terms, 1))
            backward_amplitude_0_end = s_matrix.s21 @ forward_amplitude_0_start
            forward_amplitude_0, backward_amplitude_0 = fields.colocate_amplitudes(
                forward_amplitude_start=forward_amplitude_0_start,
                backward_amplitude_end=backward_amplitude_0_end,
                z_offset=0,
                layer_solve_result=layer_solve_results[0],
                layer_thickness=thicknesses[0],
            )
            s_forward_0, s_backward_0 = fields.amplitude_poynting_flux(
                forward_amplitude_0, backward_amplitude_0, layer_solve_results[0]
            )
            reflection_s = s_backward_0[0, 0] / s_forward_0[0, 0]
            reflection_p = s_backward_0[num_terms, 0] / s_forward_0[num_terms, 0]
            onp.testing.assert_allclose(jnp.abs(reflection_s), expected_reflection_s)
            onp.testing.assert_allclose(jnp.abs(reflection_p), expected_reflection_p)

        # Check that transmission matches expected, using the poynting vector.
        with self.subTest("transmission from poynting flux"):
            forward_amplitude_1_start = s_matrix.s11 @ forward_amplitude_0_start
            backward_amplitude_1_end = jnp.zeros((2 * num_terms, 1))
            forward_amplitude_1, backward_amplitude_1 = fields.colocate_amplitudes(
                forward_amplitude_start=forward_amplitude_1_start,
                backward_amplitude_end=backward_amplitude_1_end,
                z_offset=thicknesses[1],
                layer_solve_result=layer_solve_results[1],
                layer_thickness=thicknesses[1],
            )
            s_forward_1, s_backward_1 = fields.amplitude_poynting_flux(
                forward_amplitude_1, backward_amplitude_1, layer_solve_results[1]
            )
            transmission_s = s_forward_1[0, 0] / s_forward_0[0, 0]
            transmission_p = s_forward_1[num_terms, 0] / s_forward_0[num_terms, 0]
            onp.testing.assert_allclose(
                jnp.abs(transmission_s), 1 - expected_reflection_s
            )
            onp.testing.assert_allclose(
                jnp.abs(transmission_p), 1 - expected_reflection_p
            )


class RotatedGratingTest(unittest.TestCase):
    def test_different_unit_cells_give_same_reflection(self):
        # Tests that different, nominally equivalent unit cell selections
        # actually yield equal reflectivity for a complex, wavy grating
        # structure.
        wavelength = jnp.array(0.4)
        in_plane_wavevector = jnp.array((0, 0))
        approximate_num_terms = 50
        pattern_pitch = 1.0
        pattern_width = 0.3
        pattern_modulation_depth = 0.1
        thicknesses = (0, 1, 0)

        def wavy_pattern(x, y):
            # Returns the density of a "wavy" grating at location `(x, y)`.
            xu = x % pattern_pitch  # x in the unit cell
            yu = y % pattern_pitch  # y in the unit cell
            x_start = (
                pattern_pitch / 2
                - pattern_width / 2
                + pattern_modulation_depth * jnp.cos(yu * 2 * jnp.pi)
            )
            x_end = (
                pattern_pitch / 2
                + pattern_width / 2
                + pattern_modulation_depth * jnp.cos(yu * 2 * jnp.pi)
            )
            return ((xu > x_start) & (xu < x_end)).astype(float)

        def compute_reflection(u, v):
            i, j = jnp.meshgrid(
                jnp.linspace(0, 1, 100),
                jnp.linspace(0, 1, 100),
                indexing="ij",
            )
            x = i * u[0] + j * v[0]
            y = i * u[1] + j * v[1]
            density = wavy_pattern(x, y)
            permittivities = [
                jnp.array([[1]]),
                density * 2.25 + (1 - density),
                jnp.array([[2.25]]),
            ]
            primitive_lattice_vectors = basis.LatticeVectors(
                u=jnp.asarray(u), v=jnp.asarray(v)
            )
            expansion = basis.generate_expansion(
                primitive_lattice_vectors=primitive_lattice_vectors,
                approximate_num_terms=approximate_num_terms,
                truncation=basis.Truncation.CIRCULAR,
            )
            s_matrix, layer_solve_results = _solve_s_matrix(
                wavelength=wavelength,
                in_plane_wavevector=in_plane_wavevector,
                primitive_lattice_vectors=primitive_lattice_vectors,
                expansion=expansion,
                permittivities=permittivities,
                thicknesses=thicknesses,
                formulation=fmm.Formulation.JONES_DIRECT,
            )
            rte = s_matrix.s21[0, 0]
            rtm = s_matrix.s21[expansion.num_terms, expansion.num_terms]
            return rte, rtm

        # Solve where the unit cell is square, with u=x and v=y.
        expected_rte, expected_rtm = compute_reflection(u=(1, 0), v=(0, 1))

        # Solve with a variety of other unit cells, taking care that all selections
        # have the same area. The wavevectors in the expansion are identical, but
        # there can still be slight differences in the reflection since e.g. the
        # vector field can differ.

        rte, rtm = compute_reflection(u=(0, 1), v=(1, 0))
        with self.subTest("rotated square TE"):
            onp.testing.assert_allclose(rte, expected_rte)
        with self.subTest("rotated square TM"):
            onp.testing.assert_allclose(rtm, expected_rtm)

        rte, rtm = compute_reflection(u=(1, 0), v=(1, 1))
        with self.subTest("parallelogram TE"):
            onp.testing.assert_allclose(rte, expected_rte, atol=3e-3)
        with self.subTest("parallelogram TM"):
            onp.testing.assert_allclose(rtm, expected_rtm, atol=3e-3)

        rte, rtm = compute_reflection(u=(1, 1), v=(2, 1))
        with self.subTest("skewed parallelogram TE"):
            onp.testing.assert_allclose(rte, expected_rte, atol=2e-3)
        with self.subTest("skewed parallelogram TM"):
            onp.testing.assert_allclose(rtm, expected_rtm, atol=4e-3)



"""A 1D metallic grating example, including convergence analysis.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import itertools
from typing import Tuple

import jax.numpy as jnp

from fmmax import basis, fmm, layer, scattering, utils

NUM_TERMS_SWEEP = (9, 25, 49, 81, 121, 169, 225, 289, 361, 441, 529, 625, 729, 841)


def simulate_grating(
    permittivity_ambient: complex = 1.0 + 0.0j,
    permittivity_planarization: complex = 2.25 + 0.0j,
    permittivity_substrate: complex = -7.632 + 0.731j,
    wavelength_nm: float = 500.0,
    pitch_nm: float = 180.0,
    grating_width_nm: float = 60.0,
    grating_thickness_nm: float = 80.0,
    planarization_thickness_nm: float = 20.0,
    resolution_nm: float = 1.0,
    approximate_num_terms: int = 20,
    truncation: basis.Truncation = basis.Truncation.CIRCULAR,
    formulation: fmm.Formulation = fmm.Formulation.FFT,
) -> Tuple[int, complex, complex]:
    """Computes the TE- and TM-polarized reflection from a 1D stripe grating.

    Args:
        permittivity_ambient: The permittivity of the ambient.
        permittivity_planarization: The permittivity of media encapsulating grating.
        permittivity_substrate: The permittivity of the substrate below the grating,
            and the grating itself.
        wavelength_nm: The excitation wavelength, in nanometers.
        pitch_nm: The grating pitch, in nanometers.
        grating_width_nm: The width of the lines comprising the grating.
        grating_thickness_nm: The height of the grating.
        planarization_thickness_nm: The thickness of the planarization layer above
            the grating.
        resolution_nm: The rasterization resolution for patterned layers.
        approximate_num_terms: The approximate number of terms used in the plane
            wave expansion of the fields.
        truncation: Determines the truncation of the expansion.
        formulation: Specifies the formulation to be used.

    Returns:
        The number of terms in the expansion, and the reflection coefficients for TE-
        and TM-polarization.
    """
    x_nm, _ = jnp.meshgrid(
        jnp.arange(-pitch_nm / 2, pitch_nm / 2, resolution_nm),
        jnp.arange(-pitch_nm / 2, pitch_nm / 2, resolution_nm),
        indexing="ij",
    )
    density = (jnp.abs(x_nm) <= grating_width_nm / 2).astype(float)

    permittivities = [
        jnp.asarray([[permittivity_ambient]]),
        jnp.asarray([[permittivity_planarization]]),
        utils.interpolate_permittivity(
            permittivity_solid=jnp.asarray(permittivity_substrate),
            permittivity_void=jnp.asarray(permittivity_planarization),
            density=density,
        ),
        jnp.asarray([[permittivity_substrate]]),
    ]
    thicknesses = [0, planarization_thickness_nm, grating_thickness_nm, 0]

    in_plane_wavevector = jnp.asarray([0.0, 0.0])
    primitive_lattice_vectors = basis.LatticeVectors(
        u=jnp.asarray([pitch_nm, 0.0]), v=jnp.asarray([0.0, pitch_nm])
    )
    expansion = basis.generate_expansion(
        primitive_lattice_vectors=primitive_lattice_vectors,
        approximate_num_terms=approximate_num_terms,
        truncation=truncation,
    )
    layer_solve_results = [
        layer.eigensolve_isotropic_media(
            wavelength=jnp.asarray(wavelength_nm),
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            permittivity=p,
            expansion=expansion,
            formulation=formulation,
        )
        for p in permittivities
    ]
    s_matrix = scattering.stack_s_matrix(
        layer_solve_results=layer_solve_results,
        layer_thicknesses=[jnp.asarray(t) for t in thicknesses],
    )

    r_te = s_matrix.s21[0, 0]
    r_tm = s_matrix.s21[expansion.num_terms, expansion.num_terms]
    return expansion.num_terms, complex(r_te), complex(r_tm)


def convergence_study(
    approximate_num_terms: Tuple[int, ...] = NUM_TERMS_SWEEP,
    truncations: Tuple[basis.Truncation, ...] = (
        basis.Truncation.CIRCULAR,
        basis.Truncation.PARALLELOGRAMIC,
    ),
    fmm_formulations: Tuple[fmm.Formulation, ...] = (
        fmm.Formulation.FFT,
        fmm.Formulation.JONES_DIRECT,
        fmm.Formulation.JONES,
        fmm.Formulation.NORMAL,
        fmm.Formulation.POL,
    ),
) -> Tuple[Tuple[fmm.Formulation, basis.Truncation, int, complex, complex], ...]:
    """Sweeps over number of terms and fmm formulations to study convergence."""
    results = []
    for formulation, truncation, n in itertools.product(
        fmm_formulations,
        truncations,
        approximate_num_terms,
    ):
        num_terms, r_te, r_tm = simulate_grating(
            approximate_num_terms=n,
            truncation=truncation,
            formulation=formulation,
        )
        results.append((formulation, truncation, num_terms, r_te, r_tm))
        print(
            f"{formulation.value}/{truncation.value}/n={num_terms}: "
            f"r_te={complex(r_te):.3f}, r_tm={complex(r_tm):.3f}"
        )
    return tuple(results)


if __name__ == "__main__":
    convergence_study()



"""Tests for `fmmax.beams`.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import unittest

import jax.numpy as jnp
import numpy as onp
import parameterized

from fmmax import beams


class RotatedFieldsTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            [0, 0, 0],
            [0, 0, jnp.pi / 2],
            [jnp.pi / 4, 0, 0],
            [jnp.pi / 4, jnp.pi / 4, 0],
            [jnp.pi / 4, jnp.pi / 4, jnp.pi / 2],
            [jnp.pi / 3, jnp.pi / 3, jnp.pi / 3],
        ]
    )
    def test_rotated_plane_wave_matches_expected(
        self, polar_angle, azimuthal_angle, polarization_angle
    ):
        wavelength = 0.314

        def plane_wave_fields(x, y, z):
            # Fields for an x-polarized plane wave propagating in the z-direction.
            del x, y
            ex = jnp.exp(1j * 2 * jnp.pi / wavelength * z)
            ey = jnp.zeros_like(ex)
            ez = jnp.zeros_like(ex)
            hx = jnp.zeros_like(ex)
            hy = ex
            hz = jnp.zeros_like(ex)
            return (ex, ey, ez), (hx, hy, hz)

        def rotated_plane_wave_fields(x, y, z):
            # Fields for a plane wae propagating in the direction defined by
            # `polar_angle` and `azimuthal_angle`, with polarization rotated by
            # the specified `polarization_angle`.
            kx = (
                2
                * jnp.pi
                / wavelength
                * jnp.sin(polar_angle)
                * jnp.cos(azimuthal_angle)
            )
            ky = (
                2
                * jnp.pi
                / wavelength
                * jnp.sin(polar_angle)
                * jnp.sin(azimuthal_angle)
            )
            kz = 2 * jnp.pi / wavelength * jnp.cos(polar_angle)
            exponential = jnp.exp(1j * (kx * x + ky * y + kz * z))

            # Define the two unit vectors for the polarization axes.
            pxx = jnp.cos(polar_angle) * jnp.cos(azimuthal_angle)
            pxy = jnp.cos(polar_angle) * jnp.sin(azimuthal_angle)
            pxz = -jnp.sin(polar_angle)

            pyx = -jnp.sin(azimuthal_angle)
            pyy = jnp.cos(azimuthal_angle)
            pyz = jnp.zeros_like(azimuthal_angle)

            # Check that the unit vectors are orthogonal.
            assert onp.isclose(pxx * pyx + pxy * pyy + pxz * pyz, 0.0)

            aex = pxx * jnp.cos(polarization_angle) + pyx * jnp.sin(polarization_angle)
            aey = pxy * jnp.cos(polarization_angle) + pyy * jnp.sin(polarization_angle)
            aez = pxz * jnp.cos(polarization_angle) + pyz * jnp.sin(polarization_angle)

            ahx = -pxx * jnp.sin(polarization_angle) + pyx * jnp.cos(polarization_angle)
            ahy = -pxy * jnp.sin(polarization_angle) + pyy * jnp.cos(polarization_angle)
            ahz = -pxz * jnp.sin(polarization_angle) + pyz * jnp.cos(polarization_angle)

            ex = aex * exponential
            ey = aey * exponential
            ez = aez * exponential
            hx = ahx * exponential
            hy = ahy * exponential
            hz = ahz * exponential
            return (ex, ey, ez), (hx, hy, hz)

        x, y, z = jnp.meshgrid(
            jnp.linspace(0, 1),
            jnp.linspace(0, 2),
            jnp.linspace(-1, 1),
            indexing="ij",
        )

        expected_fields = rotated_plane_wave_fields(x, y, z)
        fields = beams.shifted_rotated_fields(
            plane_wave_fields,
            x,
            y,
            z,
            beam_origin_x=0,
            beam_origin_y=0,
            beam_origin_z=0,
            polar_angle=polar_angle,
            azimuthal_angle=azimuthal_angle,
            polarization_angle=polarization_angle,
        )
        with self.subTest("ex"):
            onp.testing.assert_allclose(fields[0][0], expected_fields[0][0], atol=1e-5)
        with self.subTest("ey"):
            onp.testing.assert_allclose(fields[0][1], expected_fields[0][1], atol=1e-5)
        with self.subTest("ez"):
            onp.testing.assert_allclose(fields[0][2], expected_fields[0][2], atol=1e-5)
        with self.subTest("hx"):
            onp.testing.assert_allclose(fields[1][0], expected_fields[1][0], atol=1e-5)
        with self.subTest("hy"):
            onp.testing.assert_allclose(fields[1][1], expected_fields[1][1], atol=1e-5)
        with self.subTest("hz"):
            onp.testing.assert_allclose(fields[1][2], expected_fields[1][2], atol=1e-5)

    @parameterized.parameterized.expand(
        [
            [0, 0, 0],
            [0.5, 0, 0],
            [0, 0.5, 0],
            [0, 0, 0.5],
        ]
    )
    def test_shifts_match_expected(self, beam_origin_x, beam_origin_y, beam_origin_z):
        def field_fn(x, y, z):
            ex = jnp.exp(-(x**2) - y**2 - z**2)
            return (ex, ex, ex), (ex, ex, ex)

        def shifted_field_fn(x, y, z):
            ex = jnp.exp(
                -((x - beam_origin_x) ** 2)
                - (y - beam_origin_y) ** 2
                - (z - beam_origin_z) ** 2
            )
            return (ex, ex, ex), (ex, ex, ex)

        x, y, z = jnp.meshgrid(
            jnp.linspace(0, 1),
            jnp.linspace(0, 2),
            jnp.linspace(-1, 1),
            indexing="ij",
        )
        fields = beams.shifted_rotated_fields(
            field_fn, x, y, z, beam_origin_x, beam_origin_y, beam_origin_z, 0, 0, 0
        )
        expected = shifted_field_fn(x, y, z)
        onp.testing.assert_allclose(fields, expected)


class RotationMatrixTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            # Check that rotations move the z-axis.
            [(0, 0, 1), (0, 0, 1), 0, 0, 0],
            [(0, 0, 1), (1, 0, 0), jnp.pi / 2, 0, 0],
            [(0, 0, 1), (1 / jnp.sqrt(2), 0, 1 / jnp.sqrt(2)), jnp.pi / 4, 0, 0],
            [(0, 0, 1), (0, 1, 0), jnp.pi / 2, jnp.pi / 2, 0],
            [(1, 0, 0), (0, 1, 0), 0, 0, jnp.pi / 2],
            # Check rotations about the propagation axis.
            [(-0.1, 0, 1), (1, 0, 0.1), jnp.pi / 2, 0, 0],
            [(-0.1, 0, 1), (1, -0.1, 0.0), jnp.pi / 2, 0, jnp.pi / 2],
            [(-0.1, 0, 1), (1, 0.0, -0.1), jnp.pi / 2, 0, jnp.pi],
            [(-0.1, 0, 1), (1, 0.1, 0.0), jnp.pi / 2, 0, 3 * jnp.pi / 2],
            [(-0.1, 0, 1), (1, 0.0, 0.1), jnp.pi / 2, 0, 2 * jnp.pi],
        ]
    )
    def test_rotation_matches_expected(
        self,
        coords,
        expected_rotated_coords,
        polar_angle,
        azimuthal_angle,
        polarization_angle,
    ):
        matrix = beams.rotation_matrix(polar_angle, azimuthal_angle, polarization_angle)
        rotated_coords = matrix @ jnp.asarray(coords)[:, jnp.newaxis]
        rotated_coords = jnp.squeeze(rotated_coords, axis=-1)
        onp.testing.assert_allclose(
            rotated_coords, jnp.asarray(expected_rotated_coords), atol=1e-7
        )



"""Tests for `fmmax.utils`.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import itertools
import unittest

import jax
import jax.numpy as jnp
import numpy as onp
import parameterized

from fmmax import utils

# Enable 64-bit precision for higher accuracy.
jax.config.update("jax_enable_x64", True)

RTOL = 1e-5
RTOL_FD = 1e-3


def _jacfwd_fd(fn, delta=1e-6):
    """Forward mode jacobian by finite differences."""

    def _jac_fn(x):
        f0 = fn(x)
        jac = jnp.zeros(f0.shape + x.shape, dtype=f0.dtype)
        for inds in itertools.product(*[range(dim) for dim in x.shape]):
            offset = jnp.zeros_like(x).at[inds].set(delta)
            grad = (fn(x + offset / 2) - fn(x - offset / 2)) / delta
            jac_inds = tuple([slice(0, d) for d in f0.shape]) + inds
            jac = jac.at[jac_inds].set(grad)
        return jac

    return _jac_fn


def _sort_eigs(eigvals, eigvecs):
    """Sorts eigenvalues/eigenvectors and enforces a phase convention."""
    assert eigvals.shape[:-1] == eigvecs.shape[:-2]
    assert eigvecs.shape[-2:] == (eigvals.shape[-1],) * 2
    order = jnp.argsort(jnp.abs(eigvals), axis=-1)
    sorted_eigvals = jnp.take_along_axis(eigvals, order, axis=-1)
    sorted_eigvecs = jnp.take_along_axis(eigvecs, order[..., jnp.newaxis, :], axis=-1)
    assert eigvals.shape == sorted_eigvals.shape
    assert eigvecs.shape == sorted_eigvecs.shape
    # Set the phase of the largest component to zero.
    max_ind = jnp.argmax(jnp.abs(sorted_eigvecs), axis=-2)
    max_component = jnp.take_along_axis(
        sorted_eigvecs, max_ind[..., jnp.newaxis, :], axis=-2
    )
    sorted_eigvecs = sorted_eigvecs / jnp.exp(1j * jnp.angle(max_component))
    assert eigvecs.shape == sorted_eigvecs.shape
    return sorted_eigvals, sorted_eigvecs


def _reference_padded_conv(x, kernel, padding_mode):
    # Conservative padding equal to twice the kernel max dimension.
    p = 2 * max(kernel.shape)
    x_padded = onp.pad(x, ((p, p), (p, p)), padding_mode)
    y = jax.scipy.signal.convolve2d(x_padded, kernel[::-1, ::-1], mode="same")
    i_lo = p if kernel.shape[0] % 2 == 1 else p + 1
    j_lo = p if kernel.shape[1] % 2 == 1 else p + 1
    return y[i_lo : (i_lo + x.shape[0]), j_lo : (j_lo + x.shape[1])]


class DiagTest(unittest.TestCase):
    def test_diag_matches_expected(self):
        shapes = ((5,), (2, 5), (9, 1, 8))
        for shape in shapes:
            with self.subTest(shape):
                v = jax.random.uniform(jax.random.PRNGKey(0), shape)
                d = utils.diag(v)
                expected = jnp.zeros(shape + (shape[-1],))
                for ind in itertools.product(*[range(dim) for dim in shape[:-1]]):
                    expected = expected.at[ind].set(jnp.diag(v[ind]))
                onp.testing.assert_allclose(d, expected)


class AngularFrequencyTest(unittest.TestCase):
    def test_value_matches_expected(self):
        self.assertEqual(
            utils.angular_frequency_for_wavelength(2.71), 2 * jnp.pi / 2.71
        )


class MatrixAdjointTest(unittest.TestCase):
    def test_adjoint_matches_expected(self):
        shapes = ((5, 5), (2, 5, 5), (9, 1, 8, 8))
        for shape in shapes:
            with self.subTest(shape):
                m = jax.random.uniform(jax.random.PRNGKey(0), shape)
                ma = utils.matrix_adjoint(m)
                expected = jnp.zeros(shape)
                for ind in itertools.product(*[range(dim) for dim in shape[:-2]]):
                    expected = expected.at[ind].set(jnp.conj(m[ind]).T)
                onp.testing.assert_allclose(ma, expected)


class BatchCompatibleTest(unittest.TestCase):
    def test_value_matches_expected(self):
        shapes_and_expected = [
            ([(1, 2), (2,)], True),
            ([(1, 2), (3,)], False),
            ([(1, 2), (4, 2, 2)], True),
            ([(1, 2), ()], True),
        ]
        for shapes, expected in shapes_and_expected:
            with self.subTest(shapes):
                self.assertEqual(utils.batch_compatible_shapes(*shapes), expected)


class AtLeastNDTest(unittest.TestCase):
    def test_shape_matches_expected(self):
        shapes_n_expected = [
            [(2, 1), 1, (2, 1)],
            [(2, 1), 2, (2, 1)],
            [(2, 1), 3, (1, 2, 1)],
            [(2, 1), 4, (1, 1, 2, 1)],
        ]
        for shape, n, expected_shape in shapes_n_expected:
            with self.subTest(n):
                x = onp.zeros(shape)
                self.assertSequenceEqual(utils.atleast_nd(x, n).shape, expected_shape)


class InterpolateTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            (4.0, 2.0, 0.0, 2.0),
            (4.0, 2.0, 1.0, 4.0),
            (4.0, 2.0, 0.5, (jnp.sqrt(2) * 0.5 + jnp.sqrt(4) * 0.5) ** 2),
            (4.0 + 1.0j, 2.0, 0.0, 2.0),
            (4.0 + 1.0j, 2.0, 1.0, 4.0 + 1.0j),
            (4.0 + 1.0j, 2.0, 0.5, (jnp.sqrt(2) * 0.5 + jnp.sqrt(4 + 1.0j) * 0.5) ** 2),
        ]
    )
    def test_interpolated_matches_expected(self, p_solid, p_void, density, expected):
        result = utils.interpolate_permittivity(p_solid, p_void, density)
        onp.testing.assert_allclose(result, expected)


class EigTest(unittest.TestCase):
    def test_no_nan_gradient_with_degenerate_eigenvalues(self):
        matrix = jnp.asarray([[2.0, 0.0, 2.0], [0.0, -2.0, 0.0], [2.0, 0.0, -1.0]])
        eigval_grad = jax.grad(lambda m: jnp.sum(jnp.abs(utils.eig(m)[0])))(matrix)
        eigvec_grad = jax.grad(lambda m: jnp.sum(jnp.abs(utils.eig(m)[1])))(matrix)
        self.assertFalse(onp.any(onp.isnan(eigval_grad)))
        self.assertFalse(onp.any(onp.isnan(eigvec_grad)))

    def test_value_matches_eig_with_nondegenerate_eigenvalues(self):
        matrix = jax.random.normal(jax.random.PRNGKey(0), (2, 4, 4))
        matrix += 1j * jax.random.normal(jax.random.PRNGKey(1), (2, 4, 4))
        expected_eigval, expected_eigvec = jnp.linalg.eig(
            jax.device_put(matrix, device=jax.devices("cpu")[0])
        )
        eigval, eigvec = utils.eig(matrix)
        onp.testing.assert_array_equal(eigval, expected_eigval)
        onp.testing.assert_array_equal(eigvec, expected_eigvec)

    def test_eigvalue_jacobian_matches_expected_real_matrix(self):
        matrix = jax.random.normal(jax.random.PRNGKey(0), (2, 4, 4)).astype(complex)
        expected_jac = jax.jacrev(jnp.linalg.eigvals, holomorphic=True)(
            jax.device_put(matrix, device=jax.devices("cpu")[0])
        )
        jac = jax.jacrev(lambda x: utils.eig(x)[0], holomorphic=True)(matrix)
        onp.testing.assert_allclose(jac, expected_jac, rtol=RTOL)

    def test_eigvalue_jacobian_matches_expected_complex_matrix(self):
        matrix = jax.random.normal(jax.random.PRNGKey(0), (2, 4, 4))
        matrix += 1j * jax.random.normal(jax.random.PRNGKey(1), (2, 4, 4))
        expected_jac = jax.jacrev(jnp.linalg.eigvals, holomorphic=True)(
            jax.device_put(matrix, device=jax.devices("cpu")[0])
        )
        jac = jax.jacrev(lambda x: utils.eig(x)[0], holomorphic=True)(matrix)
        onp.testing.assert_allclose(jac, expected_jac, rtol=RTOL)

    def test_matches_eigh_hermetian_real_matrix(self):
        # Compares against `eigh`, which is valid only for Hermetian matrices. `eig`
        # and `eigh` return eigenvalues in different, random order. We must sort
        # them to facilitiate comparison.
        def _eigh(m):
            return _sort_eigs(*jnp.linalg.eigh(m, symmetrize_input=False))

        def _eig(m):
            return _sort_eigs(*utils.eig(m))

        matrix = jax.random.normal(jax.random.PRNGKey(0), (32,))
        matrix = matrix.reshape((2, 4, 4)).astype(complex)
        matrix = matrix + utils.matrix_adjoint(matrix)
        onp.testing.assert_array_equal(matrix, jnp.transpose(matrix, (0, 2, 1)))

        with self.subTest("eigenvalues"):
            onp.testing.assert_allclose(_eig(matrix)[0], _eigh(matrix)[0])

        with self.subTest("eigenvectors"):
            onp.testing.assert_allclose(_eig(matrix)[1], _eigh(matrix)[1])

        with self.subTest("eigenvalue_jac"):
            expected_eigval_jac = jax.jacrev(lambda m: _eigh(m)[0])(matrix)
            eigval_jac = jax.jacrev(lambda m: _eig(m)[0], holomorphic=True)(matrix)
            onp.testing.assert_allclose(eigval_jac, expected_eigval_jac, rtol=RTOL)

        with self.subTest("eigenvectors_jac"):
            expected_eigvec_jac = jax.jacrev(
                lambda m: _eigh(m)[1],
                holomorphic=True,
            )(matrix)
            eigvec_jac = jax.jacrev(lambda m: _eig(m)[1], holomorphic=True)(matrix)
            onp.testing.assert_allclose(eigvec_jac, expected_eigvec_jac, rtol=RTOL)

    def test_matches_eigh_hermetian_complex_matrix(self):
        # Compares against `eigh`, which is valid only for Hermetian matrices. `eig`
        # and `eigh` return eigenvalues in different, random order. We must sort
        # them to facilitiate comparison.
        def _eigh(m):
            return _sort_eigs(*jnp.linalg.eigh(m, symmetrize_input=False))

        def _eig(m):
            return _sort_eigs(*utils.eig(m))

        matrix = jax.random.normal(jax.random.PRNGKey(0), (32,))
        matrix = matrix + 1j * jax.random.normal(jax.random.PRNGKey(1), (32,))
        matrix = matrix.reshape((2, 4, 4)).astype(complex)
        matrix = matrix + utils.matrix_adjoint(matrix)
        onp.testing.assert_array_equal(matrix, utils.matrix_adjoint(matrix))

        with self.subTest("eigenvalues"):
            onp.testing.assert_allclose(_eig(matrix)[0], _eigh(matrix)[0])

        with self.subTest("eigenvectors"):
            onp.testing.assert_allclose(_eig(matrix)[1], _eigh(matrix)[1])

        with self.subTest("eigenvalues_jac"):
            expected_eigval_jac = jax.jacrev(lambda m: _eigh(m)[0])(matrix)
            eigval_jac = jax.jacrev(lambda m: _eig(m)[0], holomorphic=True)(matrix)
            onp.testing.assert_allclose(eigval_jac, expected_eigval_jac, rtol=RTOL)

        with self.subTest("eigenvectors_jac"):
            expected_eigvec_jac = jax.jacrev(
                lambda m: _eigh(m)[1],
                holomorphic=True,
            )(matrix)
            eigvec_jac = jax.jacrev(lambda m: _eig(m)[1], holomorphic=True)(matrix)
            onp.testing.assert_allclose(eigvec_jac, expected_eigvec_jac, rtol=RTOL)

    def test_eigvec_jac_matches_fd_hermetian_matrix(self):
        # Tests that a finite-difference jacobian matches that computed by the
        # custom vjp rule. Here, the input and output to the function are real,
        # but internally a complex, Hermetian matrix is passed to the
        # eigendecomposition.
        def fn(x):
            x = x + 1j * x
            x = x + utils.matrix_adjoint(x)
            _, eigvec = utils.eig(x)
            return jnp.abs(eigvec)

        matrix = jax.random.normal(jax.random.PRNGKey(0), (32,))
        matrix = matrix.reshape((2, 4, 4))

        jac = jax.jacrev(fn)(matrix)
        expected_jac = _jacfwd_fd(fn)(matrix)
        onp.testing.assert_allclose(jac, expected_jac, rtol=RTOL_FD)

    def test_eigvec_jac_matches_fd_general_complex_matrix(self):
        # Tests that a finite-difference jacobian matches that computed by the
        # custom vjp rule. Here, the input and output to the function are real,
        # but internally a complex, non-Hermetian matrix is passed to the
        # eigendecomposition.
        def fn(x):
            x = x + 1j * x
            _, eigvec = _sort_eigs(*utils.eig(x))
            return jnp.abs(eigvec)

        matrix = jax.random.normal(jax.random.PRNGKey(0), (32,))
        matrix = matrix.reshape((2, 4, 4))

        jac = jax.jacrev(fn)(matrix)
        expected_jac = _jacfwd_fd(fn)(matrix)
        onp.testing.assert_allclose(jac, expected_jac, rtol=RTOL_FD)


class MagnitudeTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            (0, 0, 0),
            (0, 1, 1),
            (1, 0, 1),
            (1, 1, jnp.sqrt(2)),
        ]
    )
    def test_magnitude(self, tx, ty, expected):
        result = utils.magnitude(tx, ty)
        onp.testing.assert_allclose(result, expected)

    def test_magnitude_gradient_no_nan(self):
        grad = jax.grad(utils.magnitude, argnums=(0, 1))(0.0, 0.0)
        self.assertFalse(onp.any(onp.isnan(grad)))


class AngleTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            (0, 0, 0),
            (0, 1, jnp.pi / 2),
            (1, 0, 0),
            (1, 1, jnp.pi / 4),
        ]
    )
    def test_angle(self, tx, ty, expected):
        result = utils.angle(tx + 1j * ty)
        onp.testing.assert_allclose(result, expected)

    def test_angle_gradient_no_nan(self):
        grad = jax.grad(utils.angle)(0.0)
        self.assertFalse(onp.any(onp.isnan(grad)))


class PaddedConvTest(unittest.TestCase):
    def test_conv_matches_reference_single(self):
        onp.random.seed(0)
        x = onp.random.rand(20, 25)
        kernel = onp.random.rand(5, 4)
        result = utils.padded_conv(x, kernel, "edge")
        expected = _reference_padded_conv(x, kernel, "edge")
        onp.testing.assert_allclose(result, expected)

    def test_conv_matches_reference_batch(self):
        onp.random.seed(0)
        x = onp.random.rand(2, 3, 20, 25)
        kernel = onp.random.rand(5, 4)

        result = utils.padded_conv(x, kernel, "edge")

        expected = onp.zeros_like(x)
        for i in range(x.shape[0]):
            for j in range(x.shape[1]):
                expected[i, j, :, :] = _reference_padded_conv(
                    x[i, j, :, :], kernel, "edge"
                )

        onp.testing.assert_allclose(result, expected)


class KernelTest(unittest.TestCase):
    def test_gaussian_kernel_matches_expected(self):
        kernel = utils.gaussian_kernel((3, 3), 2)
        expected = jnp.asarray(
            [
                [0.25, 0.50, 0.25],
                [0.50, 1.00, 0.50],
                [0.25, 0.50, 0.25],
            ]
        )
        onp.testing.assert_allclose(kernel, expected, rtol=1e-6)

    def test_gaussian_kernel_is_scale_independent(self):
        kernel_coarse = utils.gaussian_kernel((5, 7), 3)
        kernel_fine = utils.gaussian_kernel((25, 35), 15)
        onp.testing.assert_allclose(kernel_coarse, kernel_fine[2::5, 2::5], rtol=1e-6)


class ResampleTest(unittest.TestCase):
    @parameterized.parameterized.expand([[(1, 3)], [(2, 3)], [(1, 6)]])
    def test_downsampled_matches_expected(self, target_shape):
        # Downsampling where the target shape evenly divides the original
        # shape is equivalent to box downsampling.
        x = jnp.asarray(
            [[0.0, 1.0, 2.0, 3.0, 4.0, 5.0], [6.0, 7.0, 8.0, 9.0, 10.0, 11.0]],
        )
        expected = utils.box_downsample(x, target_shape)
        result = utils.resample(x, target_shape, method=jax.image.ResizeMethod.CUBIC)
        onp.testing.assert_allclose(result, expected)

    @parameterized.parameterized.expand([[(4, 12)], [(2, 12)], [(3, 9)]])
    def test_upsampled_matches_expected(self, target_shape):
        # Upsampling is equivalent to `jax.image.resize`.
        x = jnp.asarray(
            [[0.0, 1.0, 2.0, 3.0, 4.0, 5.0], [6.0, 7.0, 8.0, 9.0, 10.0, 11.0]],
        )
        expected = jax.image.resize(
            x, target_shape, method=jax.image.ResizeMethod.CUBIC
        )
        result = utils.resample(x, target_shape, method=jax.image.ResizeMethod.CUBIC)
        onp.testing.assert_allclose(result, expected)


class BoxDownsampleTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            [(4, 4), (3, 3)],
            [(4, 3), (3, 1)],
            [(4, 4, 1), (1, 1)],
        ]
    )
    def test_downsample_factor_validation(self, arr_shape, target_shape):
        with self.assertRaisesRegex(
            ValueError, "Each axis of `shape` must evenly divide "
        ):
            utils.box_downsample(jnp.ones(arr_shape), shape=target_shape)

    def test_downsampled_matches_expected(self):
        x = jnp.asarray(
            [[0.0, 1.0, 2.0, 3.0, 4.0, 5.0], [6.0, 7.0, 8.0, 9.0, 10.0, 11.0]],
        )
        expected = jnp.asarray([[14 / 4, 22 / 4, 30 / 4]])
        result = utils.box_downsample(x, (1, 3))
        onp.testing.assert_allclose(result, expected)

    def test_upsample_downsample(self):
        onp.random.seed(0)
        shape = (5, 2, 8, 3, 3)
        factor = 4
        original = jnp.asarray(onp.random.rand(*shape))

        kernel = jnp.ones((factor,) * original.ndim, dtype=original.dtype)
        upsampled = jnp.kron(original, kernel)

        expected_shape = tuple([factor * d for d in shape])
        self.assertSequenceEqual(upsampled.shape, expected_shape)
        downsampled_upsampled = utils.box_downsample(upsampled, original.shape)
        onp.testing.assert_allclose(downsampled_upsampled, original, rtol=1e-6)


class AbsoluteAxesTest(unittest.TestCase):
    @parameterized.parameterized.expand(([(0, 0), 3], [(1, -2), 3]))
    def test_absolute_axes_duplicates(self, axes, ndim):
        with self.assertRaisesRegex(ValueError, "Found duplicates in `axes`"):
            utils.absolute_axes(axes, ndim)

    @parameterized.parameterized.expand(([(3,), 3], [(-4,), 3]))
    def test_absolute_axes_out_of_range(self, axes, ndim):
        with self.assertRaisesRegex(
            ValueError, "All elements of `axes` must be in the range"
        ):
            utils.absolute_axes(axes, ndim)

    @parameterized.parameterized.expand(
        (
            [(0, 1, 2, 3), 4, (0, 1, 2, 3)],
            [(0, 1, 2, 3), 6, (0, 1, 2, 3)],
            [(-4, -3, -2, -1), 4, (0, 1, 2, 3)],
            [(-6, -5, -4, -3), 6, (0, 1, 2, 3)],
        )
    )
    def test_absolute_axes_match_expected(self, axes, ndim, expected_axes):
        absolute_axes = utils.absolute_axes(axes, ndim)
        self.assertSequenceEqual(absolute_axes, expected_axes)



"""Tests for `fmmax.layer`.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import functools
import unittest

import grcwa
import jax
import jax.numpy as jnp
import numpy as onp
import parameterized

from fmmax import basis, fmm, layer, utils

# Enable 64-bit precision for higher accuracy.
jax.config.update("jax_enable_x64", True)


WAVELENGTH = jnp.asarray(0.628)
PRIMITIVE_LATTICE_VECTORS = basis.LatticeVectors(
    u=onp.asarray((0.9, 0.1)), v=onp.asarray((0.2, 0.8))
)
EXPANSION = basis.generate_expansion(
    primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
    approximate_num_terms=4,
    truncation=basis.Truncation.CIRCULAR,
)
IN_PLANE_WAVEVECTOR = basis.plane_wave_in_plane_wavevector(
    wavelength=WAVELENGTH,
    polar_angle=jnp.pi * 0.1,
    azimuthal_angle=jnp.pi * 0.2,
    permittivity=jnp.asarray(1.0),
)


def _sort_eigs(eigvals, eigvecs):
    """Sorts eigenvalues/eigenvectors and enforces a phase convention."""
    assert eigvals.shape[:-1] == eigvecs.shape[:-2]
    assert eigvecs.shape[-2:] == (eigvals.shape[-1],) * 2
    order = jnp.argsort(jnp.abs(eigvals), axis=-1)
    sorted_eigvals = jnp.take_along_axis(eigvals, order, axis=-1)
    sorted_eigvecs = jnp.take_along_axis(eigvecs, order[..., jnp.newaxis, :], axis=-1)
    assert eigvals.shape == sorted_eigvals.shape
    assert eigvecs.shape == sorted_eigvecs.shape
    # Set the phase of the largest component to zero.
    max_ind = jnp.argmax(jnp.abs(sorted_eigvecs), axis=-2)
    max_component = jnp.take_along_axis(
        sorted_eigvecs, max_ind[..., jnp.newaxis, :], axis=-2
    )
    sorted_eigvecs = sorted_eigvecs / jnp.exp(1j * jnp.angle(max_component))
    assert eigvecs.shape == sorted_eigvecs.shape
    return sorted_eigvals, sorted_eigvecs


class GrcwaComparisonTest(unittest.TestCase):
    def test_uniform_layer_eigenvalues_and_eigenvectors(self):
        # Compares the eigenvalues and eigenvectors for a uniform layer to those
        # obtained by grcwa.
        permittivity = jnp.asarray([[3.14]])
        solve_result = layer.eigensolve_uniform_isotropic_media(
            wavelength=WAVELENGTH,
            in_plane_wavevector=IN_PLANE_WAVEVECTOR,
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
            permittivity=permittivity,
            expansion=EXPANSION,
        )

        angular_frequency = utils.angular_frequency_for_wavelength(WAVELENGTH)
        transverse_wavevectors = basis.transverse_wavevectors(
            IN_PLANE_WAVEVECTOR, PRIMITIVE_LATTICE_VECTORS, EXPANSION
        )
        kx = transverse_wavevectors[:, 0]
        ky = transverse_wavevectors[:, 1]
        (
            expected_eigenvalues,
            expected_eigenvectors,
        ) = grcwa.rcwa.SolveLayerEigensystem_uniform(
            omega=angular_frequency,
            kx=kx,
            ky=ky,
            epsilon=onp.squeeze(permittivity),
        )

        with self.subTest("eigenvalues"):
            onp.testing.assert_allclose(
                solve_result.eigenvalues**2, expected_eigenvalues**2
            )
        with self.subTest("eigenvectors"):
            onp.testing.assert_allclose(
                solve_result.eigenvectors, expected_eigenvectors
            )

    def test_patterned_layer_eigenvalues_and_eigenvectors(self):
        # Compares the eigenvalues and eigenvectors for a patterned layer to those
        # obtained by grcwa.
        permittivity = 1.0 + jax.random.uniform(jax.random.PRNGKey(0), shape=(64, 64))
        solve_result = layer.eigensolve_patterned_isotropic_media(
            wavelength=WAVELENGTH,
            in_plane_wavevector=IN_PLANE_WAVEVECTOR,
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
            permittivity=permittivity,
            expansion=EXPANSION,
            formulation=fmm.Formulation.FFT,
        )
        eigenvalues, eigenvectors = _sort_eigs(
            solve_result.eigenvalues, solve_result.eigenvectors
        )

        angular_frequency = utils.angular_frequency_for_wavelength(WAVELENGTH)
        transverse_wavevectors = basis.transverse_wavevectors(
            IN_PLANE_WAVEVECTOR, PRIMITIVE_LATTICE_VECTORS, EXPANSION
        )
        kx = transverse_wavevectors[:, 0]
        ky = transverse_wavevectors[:, 1]
        epinv, ep2 = grcwa.fft_funs.Epsilon_fft(
            dN=1 / onp.prod(permittivity.shape),
            eps_grid=permittivity,
            G=EXPANSION.basis_coefficients,
        )
        kp = grcwa.rcwa.MakeKPMatrix(
            omega=angular_frequency,
            layer_type=1,  # `1` is for patterned layers.
            epinv=epinv,
            kx=kx,
            ky=ky,
        )
        expected_eigenvalues, expected_eigenvectors = grcwa.rcwa.SolveLayerEigensystem(
            omega=angular_frequency,
            kx=kx,
            ky=ky,
            kp=kp,
            ep2=ep2,
        )
        expected_eigenvalues, expected_eigenvectors = _sort_eigs(
            expected_eigenvalues, expected_eigenvectors
        )

        with self.subTest("eigenvalues"):
            onp.testing.assert_allclose(eigenvalues**2, expected_eigenvalues**2)
        with self.subTest("eigenvectors"):
            onp.testing.assert_allclose(eigenvectors, expected_eigenvectors)


class LayerTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            (fmm.Formulation.FFT,),
            (fmm.Formulation.POL,),
            (fmm.Formulation.NORMAL,),
            (fmm.Formulation.JONES,),
            (fmm.Formulation.JONES_DIRECT,),
        ]
    )
    def test_uniform_matches_patterned(self, formulation):
        permittivity = jnp.asarray([[3.14]])
        uniform_result = layer.eigensolve_uniform_isotropic_media(
            wavelength=WAVELENGTH,
            in_plane_wavevector=IN_PLANE_WAVEVECTOR,
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
            permittivity=permittivity,
            expansion=EXPANSION,
        )
        uniform_eigenvalues, uniform_eigenvectors = _sort_eigs(
            uniform_result.eigenvalues, uniform_result.eigenvectors
        )
        patterned_result = layer.eigensolve_patterned_isotropic_media(
            wavelength=WAVELENGTH,
            in_plane_wavevector=IN_PLANE_WAVEVECTOR,
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
            permittivity=jnp.broadcast_to(permittivity, (64, 64)),
            expansion=EXPANSION,
            formulation=formulation,
        )
        patterned_eigenvalues, patterned_eigenvectors = _sort_eigs(
            patterned_result.eigenvalues, patterned_result.eigenvectors
        )
        with self.subTest("eigenvalues"):
            onp.testing.assert_allclose(
                uniform_eigenvalues**2, patterned_eigenvalues**2
            )
        with self.subTest("z_permittivity_matrix"):
            onp.testing.assert_allclose(
                uniform_result.z_permittivity_matrix,
                patterned_result.z_permittivity_matrix,
            )
        with self.subTest("omega_script_k_matrix"):
            onp.testing.assert_allclose(
                uniform_result.omega_script_k_matrix,
                patterned_result.omega_script_k_matrix,
            )

    def test_uniform_layer_batch_matches_single(self):
        wavelength = jnp.asarray([[[0.2]], [[0.3]], [[0.4]]])
        in_plane_wavevector = jnp.asarray([[[0.0, 0.0]], [[0.05, 0.08]]])
        permittivity = jnp.asarray([[[3.14]], [[6.28]]])
        result = layer.eigensolve_uniform_isotropic_media(
            wavelength=wavelength,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
            permittivity=permittivity,
            expansion=EXPANSION,
        )
        for i, w in enumerate(wavelength[:, 0, 0]):
            for j, ipwv in enumerate(in_plane_wavevector[:, 0, :]):
                for k, p in enumerate(permittivity[:, :, :]):
                    single_result = layer.eigensolve_uniform_isotropic_media(
                        wavelength=w,
                        in_plane_wavevector=ipwv,
                        primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
                        permittivity=p,
                        expansion=EXPANSION,
                    )
                    onp.testing.assert_array_equal(
                        result.eigenvalues[i, j, k, :],
                        single_result.eigenvalues,
                    )
                    onp.testing.assert_array_equal(
                        result.eigenvectors[i, j, k, :, :],
                        single_result.eigenvectors,
                    )
                    onp.testing.assert_array_equal(
                        result.z_permittivity_matrix[i, j, k, :, :],
                        single_result.z_permittivity_matrix,
                    )
                    onp.testing.assert_array_equal(
                        result.omega_script_k_matrix[i, j, k, :, :],
                        single_result.omega_script_k_matrix,
                    )

    @parameterized.parameterized.expand(
        [
            (fmm.Formulation.FFT,),
            (fmm.Formulation.POL,),
            (fmm.Formulation.NORMAL,),
            (fmm.Formulation.JONES,),
            (fmm.Formulation.JONES_DIRECT,),
        ]
    )
    def test_patterned_layer_batch_matches_single(self, formulation):
        wavelength = jnp.asarray([[[0.2]], [[0.3]], [[0.4]]])
        in_plane_wavevector = jnp.asarray([[[0.0, 0.0]], [[0.05, 0.08]]])
        permittivity = jnp.asarray(
            [
                1.0 + jax.random.uniform(jax.random.PRNGKey(0), shape=(64, 60)),
                1.0 + jax.random.uniform(jax.random.PRNGKey(1), shape=(64, 60)),
            ]
        )
        result = layer.eigensolve_patterned_isotropic_media(
            wavelength=wavelength,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
            permittivity=permittivity,
            expansion=EXPANSION,
            formulation=formulation,
        )
        for i, w in enumerate(wavelength[:, 0, 0]):
            for j, ipwv in enumerate(in_plane_wavevector[:, 0, :]):
                for k, p in enumerate(permittivity[:, :, :]):
                    single_result = layer.eigensolve_patterned_isotropic_media(
                        wavelength=w,
                        in_plane_wavevector=ipwv,
                        primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
                        permittivity=p,
                        expansion=EXPANSION,
                        formulation=formulation,
                    )
                    eigenvalues, eigenvectors = _sort_eigs(
                        result.eigenvalues[i, j, k, :],
                        result.eigenvectors[i, j, k, :, :],
                    )
                    expected_eigenvalues, expected_eigenvectors = _sort_eigs(
                        single_result.eigenvalues, single_result.eigenvectors
                    )
                    onp.testing.assert_allclose(
                        eigenvalues**2, expected_eigenvalues**2
                    )
                    onp.testing.assert_allclose(
                        result.z_permittivity_matrix[i, j, k, :, :],
                        single_result.z_permittivity_matrix,
                    )
                    onp.testing.assert_allclose(
                        result.omega_script_k_matrix[i, j, k, :, :],
                        single_result.omega_script_k_matrix,
                    )

    def test_shape_insufficient_validation(self):
        min_shape = fmm._min_array_shape_for_expansion(EXPANSION)
        permittivity = jnp.ones(tuple([d - 1 for d in min_shape]))
        with self.assertRaisesRegex(ValueError, "`shape` is insufficient for"):
            layer.eigensolve_patterned_isotropic_media(
                wavelength=WAVELENGTH,
                in_plane_wavevector=IN_PLANE_WAVEVECTOR,
                primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
                permittivity=permittivity,
                expansion=EXPANSION,
                formulation=fmm.Formulation.FFT,
            )


class JitTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            (fmm.Formulation.FFT,),
            (fmm.Formulation.POL,),
        ]
    )
    def test_can_jit(self, formulation):
        wavelength = jnp.asarray([[[0.2]], [[0.3]], [[0.4]]])
        in_plane_wavevector = jnp.asarray([[[0.0, 0.0]], [[0.05, 0.08]]])
        permittivity = jnp.asarray(
            [
                1.0 + jax.random.uniform(jax.random.PRNGKey(0), shape=(64, 60)),
                1.0 + jax.random.uniform(jax.random.PRNGKey(1), shape=(64, 60)),
            ]
        )
        result = layer.eigensolve_patterned_isotropic_media(
            wavelength=wavelength,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
            permittivity=permittivity,
            expansion=EXPANSION,
            formulation=formulation,
        )
        jit_fn = jax.jit(
            functools.partial(
                layer.eigensolve_patterned_isotropic_media,
                expansion=EXPANSION,
            )
        )
        jit_result = jit_fn(
            wavelength=wavelength,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
            permittivity=permittivity,
            formulation=formulation,
        )
        onp.testing.assert_allclose(
            result.eigenvalues**2, jit_result.eigenvalues**2
        )


class AnistropicLayerTest(unittest.TestCase):
    def test_compare_when_layer_is_isotropic(self):
        wavelength = jnp.asarray([[[0.2]], [[0.3]], [[0.4]]])
        in_plane_wavevector = jnp.asarray([[[0.0, 0.0]], [[0.05, 0.08]]])
        permittivity = jnp.asarray(
            [
                1.0 + jax.random.uniform(jax.random.PRNGKey(0), shape=(64, 60)),
                1.0 + jax.random.uniform(jax.random.PRNGKey(1), shape=(64, 60)),
            ]
        )
        expected = layer.eigensolve_patterned_isotropic_media(
            wavelength=wavelength,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
            permittivity=permittivity,
            expansion=EXPANSION,
            formulation=fmm.Formulation.FFT,
        )
        result = layer.eigensolve_patterned_anisotropic_media(
            wavelength=wavelength,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
            permittivity_xx=permittivity,
            permittivity_xy=jnp.zeros_like(permittivity),
            permittivity_yx=jnp.zeros_like(permittivity),
            permittivity_yy=permittivity,
            permittivity_zz=permittivity,
            expansion=EXPANSION,
            formulation=fmm.Formulation.FFT,
        )
        onp.testing.assert_allclose(result.eigenvalues**2, expected.eigenvalues**2)

    def test_uniform_matches_patterned(self):
        permittivity_xx = jnp.asarray([[2.0]])
        permittivity_xy = jnp.asarray([[0.1]])
        permittivity_yx = jnp.asarray([[0.2]])
        permittivity_yy = jnp.asarray([[2.5]])
        permittivity_zz = jnp.asarray([[3.0]])
        uniform_result = layer.eigensolve_uniform_anisotropic_media(
            wavelength=WAVELENGTH,
            in_plane_wavevector=IN_PLANE_WAVEVECTOR,
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
            permittivity_xx=permittivity_xx,
            permittivity_xy=permittivity_xy,
            permittivity_yx=permittivity_yx,
            permittivity_yy=permittivity_yy,
            permittivity_zz=permittivity_zz,
            expansion=EXPANSION,
        )
        uniform_eigenvalues, uniform_eigenvectors = _sort_eigs(
            uniform_result.eigenvalues, uniform_result.eigenvectors
        )
        patterned_result = layer.eigensolve_patterned_anisotropic_media(
            wavelength=WAVELENGTH,
            in_plane_wavevector=IN_PLANE_WAVEVECTOR,
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
            permittivity_xx=jnp.broadcast_to(permittivity_xx, (64, 64)),
            permittivity_xy=jnp.broadcast_to(permittivity_xy, (64, 64)),
            permittivity_yx=jnp.broadcast_to(permittivity_yx, (64, 64)),
            permittivity_yy=jnp.broadcast_to(permittivity_yy, (64, 64)),
            permittivity_zz=jnp.broadcast_to(permittivity_zz, (64, 64)),
            expansion=EXPANSION,
            formulation=fmm.Formulation.FFT,
        )
        patterned_eigenvalues, patterned_eigenvectors = _sort_eigs(
            patterned_result.eigenvalues, patterned_result.eigenvectors
        )
        with self.subTest("eigenvalues"):
            onp.testing.assert_allclose(
                uniform_eigenvalues**2, patterned_eigenvalues**2
            )
        with self.subTest("z_permittivity_matrix"):
            onp.testing.assert_allclose(
                uniform_result.z_permittivity_matrix,
                patterned_result.z_permittivity_matrix,
            )
        with self.subTest("omega_script_k_matrix"):
            onp.testing.assert_allclose(
                uniform_result.omega_script_k_matrix,
                patterned_result.omega_script_k_matrix,
            )


class UtilityFunctionTests(unittest.TestCase):
    def test_select_sign(self):
        eigenvalues = jnp.array(
            [1 + 1e-7j, 1 - 1e-7j, 1e-3 + 1e-7j, 1e-3 - 1e-7j, 1e-7 - 1e-3j]
        )
        expected = jnp.array(
            [1 + 1e-7j, -1 + 1e-7j, 1e-3 + 1e-7j, -1e-3 + 1e-7j, -1e-7 + 1e-3j]
        )
        result = layer._select_eigenvalues_sign(eigenvalues)
        onp.testing.assert_array_equal(result, expected)


class LayerSolveResultInputValidationTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            ("wavelength", jnp.ones((1,))),
            ("in_plane_wavevector", jnp.ones((1,))),
            ("eigenvalues", jnp.ones((1,))),
            ("eigenvectors", jnp.ones((3, 4, 5, 1, 1))),
            ("eta_matrix", jnp.ones((1,))),
            ("z_permittivity_matrix", jnp.ones((1,))),
            ("omega_script_k_matrix", jnp.ones((1,))),
        ]
    )
    def test_invalid_shape(self, name, invalid_shape):
        expansion = basis.generate_expansion(
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
            approximate_num_terms=20,
            truncation=basis.Truncation.CIRCULAR,
        )
        num = expansion.num_terms
        kwargs = {
            "wavelength": jnp.ones((3, 1, 1)),
            "in_plane_wavevector": jnp.ones((1, 4, 5, 2)),
            "primitive_lattice_vectors": PRIMITIVE_LATTICE_VECTORS,
            "expansion": expansion,
            "eigenvalues": jnp.ones((3, 4, 5, 2 * num)),
            "eigenvectors": jnp.ones((3, 4, 5, 2 * num, 2 * num)),
            "eta_matrix": jnp.ones((3, 4, 5, num, num)),
            "z_permittivity_matrix": jnp.ones((3, 4, 5, num, num)),
            "omega_script_k_matrix": jnp.ones((3, 4, 5, 2 * num, 2 * num)),
        }
        kwargs[name] = invalid_shape
        with self.assertRaisesRegex(ValueError, f"`{name}` must have "):
            layer.LayerSolveResult(**kwargs)
